// Full-text search demo. (c) Peter Kankowski, 2014  http://www.strchr.com

#define _CRT_SECURE_NO_WARNINGS
#ifdef _MSC_VER
#include <crtdbg.h>
#endif

#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// ==============
// Data structure

char * g_content = NULL; // The file content

const char ** g_index = NULL; // An index for the full-text search
size_t g_index_len = 0; // The number of words in the index

// ==============
// Error handling

// sysexits
#define EX_USAGE 64
#define EX_NOINPUT 66
#define EX_OSERR 71

// Similar to err() in Linux/FreeBSD
void handle_error(int status, const char * format, ...) {
	va_list args;
	
	va_start(args, format);
	vfprintf(stderr, format, args);
	va_end(args);
	
	fprintf(stderr, "%s\n", strerror(errno));

	free(g_content);
	free(g_index);
	
	exit(status);
}

// ===================
// Construct the index

// Add a new word to the index
void add_word(const char * word) {
	// The index is reallocated after each INDEX_CHUNK elements added to it
	const size_t INDEX_CHUNK = 1024;
	
	if (g_index_len % INDEX_CHUNK == 0) {
		const char ** new_index = (const char **)realloc(g_index,
							(g_index_len + INDEX_CHUNK) * sizeof(char *));
		if (!new_index) {
			free(g_index);
			free(g_content);
			handle_error(EX_OSERR, "Out of memory (%lu words requested)\n",
						 (unsigned long)(g_index_len + INDEX_CHUNK));
		}
		g_index = new_index;
	}
	g_index[g_index_len++] = word;
}

void read_file(const char * file_name) {
	FILE * f = fopen(file_name, "rb");
	if (!f)
		handle_error(EX_NOINPUT, "Unable to open file %s\n", file_name);
	
	// Get file size
	long file_size = 0;
	if (0 != fseek(f, 0, SEEK_END) ||
		(file_size = ftell(f)) == -1 ||
		0 != fseek(f, 0, SEEK_SET)) {
		handle_error(EX_OSERR, "Get file size failed\n");
	}
	
	// Allocate memory
	g_content = (char *)malloc(file_size + 1);
	if (!g_content)
		handle_error(EX_OSERR, "Out of memory (%ld characters requested)\n", file_size);
	
	// Read the file
	size_t content_len = fread(g_content, 1, file_size, f);
	g_content[content_len] = '\0';
	
	if (content_len != (size_t)file_size)
		handle_error(EX_OSERR, "Read failed (%ld characters requested, %lu found)\n",
					 file_size, (unsigned long)content_len);
	
	fclose(f);
}

// Comparison function for qsort
int compare_strings (const void * a_ptr, const void * b_ptr) {
	const char ** a = (const char **)a_ptr;
	const char ** b = (const char **)b_ptr;
	return strcmp(*a, *b);
}

void construct_index(const char * file_name) {
	read_file(file_name);
	
	const char * word = g_content;
	
	// Word is defined as anything between whitespace characters
	
	// Skip any whitespace at the beginning of the file
	while (isspace(*word))
		word++;
	
	while (*word != '\0') {
		// Add a new word
		add_word(word);
		
		// Skip its letters
		do {
			word++;
			if (*word == '\0')
				return;
		} while (!isspace(*word));
		
		// Skip the whitespace
		do
			word++;
		while (isspace(*word));
	}
	
	// Sort the offsets by words
	qsort(g_index, g_index_len, sizeof(char *), compare_strings);
}

// ===================
// Search in the index

// Comparison function for qsort
int compare_pointers (const void * a_ptr, const void * b_ptr) {
	const char ** a = (const char **)a_ptr;
	const char ** b = (const char **)b_ptr;
	
	// Don't use subtraction (*a - *b),
	// because the pointer size may be greater than the integer size
	
	return (*a > *b) - (*a < *b);
}

// Returns the array of matches
char ** search_for_pattern(const char * pattern) {
	size_t pattern_len = strlen(pattern);
	if (pattern_len == 0)
		return NULL;
	
	// Binary search in the index (find the first match)
	size_t lo = (size_t)-1;
	size_t hi = g_index_len;
	
	while (lo + 1 < hi) {
		size_t mid = (lo + hi) / 2;
		
		if (strncmp(g_index[mid], pattern, pattern_len) < 0)
			lo = mid;
		else
			hi = mid;
	}
	
	// Determine the number of matches
	lo = hi;
	while (hi < g_index_len && strncmp(g_index[hi], pattern, pattern_len) == 0) {
		hi++;
	}
	
	// Copy the matches into a temporate array
	char ** found = (char **)malloc((hi - lo + 1) * sizeof(char *));
	if (!found)
		handle_error(EX_OSERR, "Out of memory (%lu words requested)\n", (unsigned long)(hi - lo + 1));
	memcpy(found, g_index + lo, (hi - lo) * sizeof(char *));
	found[hi - lo] = NULL;
	
	// Sort the matches by their position in the source file
	qsort(found, hi - lo, sizeof(char *), compare_pointers);
	return found;
}

// =============
// Main function

int main(int argc, char ** argv) {

#if defined(_DEBUG) && defined(_MSC_VER)
	// Detect memory leaks
	_CrtSetDbgFlag(
		_CrtSetDbgFlag(_CRTDBG_REPORT_FLAG) |
		_CRTDBG_CHECK_ALWAYS_DF |
		_CRTDBG_LEAK_CHECK_DF
	);
#endif
	if (argc == 1)
		handle_error(EX_USAGE, "Usage: %s text_file_name\n", argv[0]);
	
	construct_index(argv[1]);
	
	printf("Full-text search demo. http://www.strchr.com\n");
	printf("Enter a phrase to search for or an empty line to exit:\n");
	
	char pattern[1024];
	while (fgets(pattern, sizeof(pattern), stdin) != NULL) {
		// Trim the newline
		char * last_char = strchr(pattern, '\0') - 1;
		if (last_char >= pattern && *last_char == '\n')
			*last_char-- = '\0';
		if (last_char >= pattern && *last_char == '\r')
			*last_char-- = '\0';
		
		if (last_char < pattern) // An empty line
			break;
		
		// Run the search and display the results
		char ** found = search_for_pattern (pattern);
		
		char ** item = found;
		while (*item) {
			printf("=> %.70s\n", *item);
			item++;
		}
		
		free(found);
		printf("\n");
	}

	free(g_content);
	free(g_index);
	return 0;
}
